import os
import ast
from scrapy.http import HtmlResponse, Request


def response_from_file(filename, url=None):
    """
    Creates a HtmlResponse object based in the content of a html file
    :param filename: filename or path to a html file
    :return: HtmlResponse object with the data fetched from a file
    """
    if not url:
        url = 'http://www.fakeurl.com'

    request = Request(url=url)
    if not filename[0] == '/':
        responses_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "htmls"))
        file_path = os.path.join(responses_dir, filename)
    else:
        file_path = filename

    file_content = open(file_path, 'r').read()

    response = HtmlResponse(
        url=url, request=request, body=file_content, encoding='utf-8')
    return response


def read_results_from_file(filename):
    """
    Reads results saved in a file and returns an array with a dictionary
    (if the result is a valid json) or string otherwise
    :param filename: file name where results are stored
    :return: list of results, which may be either str or dict
    """
    if not filename[0] == '/':
        responses_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "scraped_data"))
        file_path = os.path.join(responses_dir, filename)
    else:
        file_path = filename

    arr = []
    with open(file_path, 'r') as f:
        for line in f:
            try:
                item = ast.literal_eval(line)
            except (ValueError, SyntaxError):
                item = str(line)
            arr.append(item)

    return arr



from scrapy import Request
from scraper_factory.core.utils import remove_query_string
from scraper_factory.core.base_spider import BaseSpider


class AmazonWishlistSpider(BaseSpider):
    metadata = {
        'name': 'amazon-wishlist',
        'version': '0.1.0',
        'description': 'Scrapes an amazon wishlist and returns all '
                       'items in json format',
        'parameters': [
            {
                'name': 'url',
                'type': 'str',
                'default': None,
                'required': True,
                'description': 'Amazon wishlist url to be scraped'
            }, {
                'name': 'detailed',
                'type': 'bool',
                'default': False,
                'required': False,
                'description': 'Get detailed info of each item'
            }
        ]
    }

    def __init__(self, url, queue, detailed=False, **kwargs):
        super().__init__(self.metadata['name'], url, queue, **kwargs)
        self.__detailed = detailed

    def __get_sold_by(self, sold_by_list):
        sold_by = ''
        for el in sold_by_list:
            el = el.strip()
            if any([k in el for k in ('execute', 'function')]):
                continue
            if el == '.':
                sold_by = sold_by[:-1] + '.'
                break
            sold_by += el + ' '
        return sold_by.rstrip()

    def __get_prices(self, prices_list):
        list_price = None
        discount = None
        for el in prices_list:
            if el[0] == '$':
                list_price = el.strip()
            if el.find('$') > 0:
                discount = el[el.find('$'):].strip()
        return list_price, discount

    def __parse_item(self, response, obj={}):
        stock = response.css('#availability').css('span::text').extract_first()
        sold_by_list = response.css('#merchant-info *::text').extract()
        sold_by = self.__get_sold_by(sold_by_list)
        prices_list = response.css('#buyBoxInner') \
            .css('span.a-color-secondary::text').extract()
        list_price, discount = self.__get_prices(prices_list)

        obj['stock'] = stock.strip() if stock else None
        obj['sold_by'] = sold_by if sold_by else None
        obj['list_price'] = list_price
        obj['discount'] = discount
        self.q.put(obj)
        yield obj

    def parse(self, response):
        page_items = response.css('.g-item-sortable')

        for item in page_items:
            id = item.css('li::attr(data-itemid)').extract_first()
            title = item.css('#itemName_' + id + '::text').extract_first()
            link = item.css('#itemName_' + id + '::attr(href)')\
                .extract_first()
            if link:
                link = self.base_url + link
            img = item.css('#itemImage_' + id).css('img::attr(src)')\
                .extract_first()
            byline = item.css('#item-byline-' + id + '::text').extract_first()
            price = item.css('#itemPrice_' + id).css('span::text')\
                .extract_first()

            obj = {
                'id': id,
                'title': title,
                'byline': byline,
                'price': price,
                'link': remove_query_string(link),
                'img': remove_query_string(img)
            }

            # if detailed is True, scrapes each
            # wishlist item web page to get detailed info
            if self.__detailed:
                yield Request(link,
                              callback=self.__parse_item,
                              cb_kwargs={'obj': obj})
            else:
                self.q.put(obj)
                yield obj

        # manage "infinite scrolldown"
        has_next = response.css('#sort-by-price-next-batch-lek'
                                '::attr(value)').extract_first()
        if has_next:
            lek_uri = response.css(
                '#sort-by-price-load-more-items-url-next-batch::attr(value)')\
                .extract_first()
            next_page = self.base_url + lek_uri
            yield Request(next_page)



from unittest import TestCase
from tests.utils import read_from_file
from scraper_factory.core.api import scrape


class SpiderTestBase(TestCase):

    def verify_url_results(self, spider, url, results_file, detailed=False):
        results = scrape(spider, url=url, detailed=detailed)
        correct_results = read_from_file.read_results_from_file(results_file)

        self.assertEqual(len(list(results)), len(correct_results))

        sorted_results = []
        for item in results:
            id = item['id']
            for i in range(len(correct_results)):
                if correct_results[i]['id'] != id:
                    continue
                sorted_results.append(correct_results[i])
                correct_results.pop(i)
                break

        for (test_result, proper_result) in zip(results, sorted_results):
            # when items are no longer available on Amazon they don't have
            # price anymore and it's expected that eventually, the proper
            # result price will be 'None' so they all have been set
            # accordingly but if the item does have a price, only its first
            # character will be compared with the expected '$' one
            for key in ('price', 'list_price', 'discount'):
                value = test_result.get(key)
                if value is not None:
                    test_result[key] = value[0]
                    proper_result[key] = '$'

            for key in ('sold_by', 'stock'):
                value = test_result.get(key)
                if value is not None:
                    test_result[key] = None

            self.assertDictEqual(test_result, proper_result)



from scrapy import crawler


def err_callback(failure, err_q):
    """
    Error callback, called upon errors on crawler
    :param failure: twisted.Failure object
    :param err_q: queue where errors are handed to parent process
    :return: Nothing
    """
    err_q.put(failure.value)


def run_spider(spider_cls, err_queue, **kwargs):
    """
    Scrapes a url with the given spider class
    :param spider_cls: spider constructor to be used
    :param url: string with the url to scrape
    :param data_queue: multiprocessing Queue where data will be stored
    :param err_queue: multiprocessing Queue where errors will be stored
    :return: Nothing
    """
    process = crawler.CrawlerProcess(settings={
        'FEED_FORMAT': 'json',
        'LOG_LEVEL': 'ERROR',
        'USER_AGENT': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:38.0) '
                      'Gecko/20100101 Firefox/38.0'
    })
    p = process.crawl(crawler_or_spidercls=spider_cls, **kwargs)
    p.addErrback(err_callback, err_queue)
    process.start()



from multiprocessing import Process, Queue
from scraper_factory.core import utils
from scraper_factory.core.scraper.scrapy import run_spider
from scraper_factory.core.spidermanager import SpiderManager
from scraper_factory.core.exceptions import SpiderNotFoundError


def scrape(spider_name, **kwargs):
    m = SpiderManager()
    spider_cls = m.instance(spider_name)

    if not spider_cls:
        msg = 'Spider {} doesn\'t exist'.format(spider_name)
        raise SpiderNotFoundError(msg)

    err_q = Queue()
    data_q = Queue()
    kwargs['queue'] = data_q

    p = Process(target=run_spider,
                args=(spider_cls, err_q),
                kwargs=kwargs)
    p.start()
    p.join()

    if not err_q.empty():
        raise err_q.get()

    return utils.queue_to_list(data_q)


def spiders():
    m = SpiderManager()
    return m.get_spiders()



import os
import re
import ast
import importlib.util
from scraper_factory.core import base_spider

HERE = os.path.abspath(os.path.dirname(__file__))
DEFAULT_PATH = os.path.join(HERE, os.path.pardir, 'spiders')


class SpiderManager(object):
    VALID_FILE_REGEX = r'(([a-zA-Z0-9]+)|-|_)+[^__]\.py'

    def __init__(self):
        path = os.getenv('SPIDER_PATH', DEFAULT_PATH)
        if not os.path.isdir(path):
            raise AttributeError('Invalid path {}'.format(path))

        self.spiders = {}
        self.load(path)

    def load(self, path):
        """
        Loads all the valid spiders found in the path received
        :param path: string with the path to a folder where spiders are stored
        :return: Nothing
        """
        self.spiders = {}
        for f in os.listdir(path):
            if not re.match(self.VALID_FILE_REGEX, f):
                continue

            spider_file = os.path.join(path, f)
            try:
                sp = self.__load_spider_from_file(spider_file)
                metadata = sp.metadata
                self.spiders[metadata['name']] = sp.metadata
                self.spiders[metadata['name']]['instance'] = sp
            except (AttributeError, ImportError) as e:
                print('Skipped {}: {}'.format(f, e))
                continue

    def __load_spider_from_file(self, filename):
        """
        Loads a spider from a certain file. The file must contain
        only one class, derived from BaseSpider
        :param filename: string containing the path to the definition
        of a spider
        :return: constructor of the class if it was properly loaded,
        None otherwise
        """
        with open(filename) as f:
            node = ast.parse(f.read())

        classes = [n for n in node.body if isinstance(n, ast.ClassDef)]
        if len(classes) > 1:
            raise AttributeError('File can\'t contain more than one class')
        spider_name = classes[0].name

        spec = importlib.util.spec_from_file_location(
            spider_name, filename)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        self.__validate_module(module, spider_name)

        return getattr(module, spider_name)

    def __validate_module(self, module, classname):
        """
        Validates that the module received is a valid spider definition
        :param module: python module
        :return: Nothing. Raises exception on failure.
        """
        cls = getattr(module, classname)
        base_spider.BaseSpider.check_metadata(cls.metadata)

        if not issubclass(cls, base_spider.BaseSpider):
            msg = '{} found doesn\'t derive from BaseSpider'.format(classname)
            raise ImportError(msg)

        base_spider.BaseSpider.check_metadata(cls.metadata)

    def instance(self, name):
        """
        Returns the constructor to the class with a given name.
        :param name: string with the spider name
        : return: constructor of the desired class. None on error.
        """
        sp = self.spiders.get(name)
        if not sp:
            return None

        return sp.get('instance')

    def get_spiders(self):
        """
        Returns a list with the metadata of all available spiders
        :return: list of dicts containing spider metadata
        """
        arr = []
        for name, metadata in self.spiders.items():
            arr.append(
                {k: v for k, v in metadata.items() if k not in 'instance'})
        return arr



from multiprocessing.queues import Queue
from urllib.parse import urlparse


def queue_to_list(q):
    """
    Transforms a multiprocessing queue into an array
    :param q: multiprocessing Queue object
    :return: list with the elements from the queue
    """
    if not isinstance(q, Queue):
        raise TypeError('Argument must be a multiprocessing Queue')

    arr = []
    while not q.empty():
        arr.append(q.get())
    return arr


def remove_query_string(url):
    """
    Removes query string from a url
    :param url: string with a url
    :return: clean base url
    """
    if not isinstance(url, str):
        raise TypeError('Argument must be a string')
    return url.split('?')[0]


def validate_url(url):
    """
    Checks that the given string contains valid url
    :param url: string with a url
    :return: True if url is valid, False otherwise
    """
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except (ValueError, TypeError, AttributeError):
        return False



import os
import codecs
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


def read(*folders):
    with codecs.open(os.path.join(here, *folders), encoding='utf-8') as f:
        return f.read()


def get_requirements(file_name):
    requires_file = read('requirements', file_name)
    return requires_file.splitlines()


long_description = read('README.rst')

setup(
    name='scraper-factory',

    version='0.2.0',

    description='Scraping library to retrieve data from useful pages, such as Amazon wishlists',
    long_description=long_description,

    url='https://github.com/machinia/scraper-factory',

    author='Pablo Ahumada, Jorge Capona',
    author_email='pablo.ahumadadiaz@gmail.com, jcapona@gmail.com',

    license='MIT',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: System',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='scraping wishlist amazon',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=get_requirements('default.txt'),
    extras_require={},
    package_data={},
    data_files=[],
    entry_points={},
    test_suite='tests',
    setup_requires=get_requirements('tests.txt'),
)


from abc import ABC, abstractmethod
from urllib.parse import urlparse
from scrapy import Spider
from scraper_factory.core.utils import validate_url
from scraper_factory.core.exceptions import InvalidUrlError


class BaseSpider(ABC, Spider):
    __METADATA_KEYS = ('name', 'version', 'description', 'parameters')
    metadata = {}

    def __init__(self, name, uri, queue, **kwargs):
        if not validate_url(uri):
            raise InvalidUrlError('Invalid URL')

        self.check_metadata(self.metadata)

        self.name = name
        parsed_url = urlparse(uri)
        self.start_urls = [uri]
        self.base_url = parsed_url.scheme + '://' + parsed_url.netloc
        self.allowed_domains = []
        self.allowed_domains.append(parsed_url.netloc)
        self.allowed_domains.append(uri.split('//')[-1])

        self.q = queue
        super().__init__(**kwargs)

    @abstractmethod
    def parse(self, response):
        """This method should implement how to parse the
         response data."""

    @staticmethod
    def check_metadata(metadata):
        for key in BaseSpider.__METADATA_KEYS:
            if not metadata.get(key):
                msg = '"{}" not defined in metadata'.format(key)
                raise AttributeError(msg)
